@@ -444,6 +444,9 @@ class ExoRPC():
             r.append(response)
         return r
 
+    def mult(self, cik, commands):
+        return _exomult(cik, commands)
+
     def _exomult(self, cik, commands):
         '''Takes a list of onep commands with cik omitted, e.g.:
             [['info', {alias: ""}], ['listing']]'''
