@@ -18,6 +18,7 @@ import itertools
 import os
 
 import six
+import yaml
 from six import iteritems
 from dateutil import parser
 
@@ -238,10 +239,6 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
             cik, 'dataport', {'format': 'integer', 'name': 'int_port'},
             write=['-1', '0', '100000000'],
             record=[[665366400, '42']])
-        stdports['boolean'] = Resource(
-            cik, 'dataport', {'format': 'boolean', 'name': 'boolean_port'},
-            write=['false', 'true', 'false'],
-            record=[[-100, 'true'], [-200, 'false'], [-300, 'true']])
         stdports['string'] = Resource(
             cik, 'dataport', {'format': 'string', 'name': 'string_port'},
             alias='string_port_alias',
@@ -252,7 +249,6 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
             write=['-0.1234567', '0', '3.5', '100000000.1'],
             record=[[-100, '-0.1234567'], [-200, '0'], [-300, '3.5'], [-400, '10000000.1']])
             # TODO: handle scientific notation from OneP '-0.00001'
-        # TODO: handle binary dataport
 
         self._createMultiple(cik, list(stdports.values()))
 
@@ -414,25 +410,31 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         # call did not fail
         self.ok(r, 'tree shouldn\'t fail')
         # starts with cik
-        self.l(r.stdout)
         self.assertTrue(
             re.match(".* cik: {0}.*".format(cik), r.stdout) is not None)
 
-        # has correct number of lines
-        self.assertTrue(len(r.stdout.split('\n')) == len(stdports) + 1 + 1)
+        delim = '\n'.encode('utf-8')
+        def get_lines(stdout):
+            # has correct number of lines
+            if sys.version_info < (3, 0):
+                so = stdout
+            else:
+                so = stdout.encode('utf-8')
+            return so.split(delim)
+
+        self.assertTrue(len(get_lines(r.stdout)) == len(stdports) + 1 + 1)
 
         r = rpc('tree', cik, '--level=0')
         self.ok(r, 'tree with --level=0 shouldn\'t fail')
-        self.ok(r)
-        self.assertTrue(len(r.stdout.split('\n')) == 1)
+        self.assertTrue(len(get_lines(r.stdout)) == 1)
 
         r = rpc('tree', cik, '--level=1')
         self.ok(r, 'tree with --level=1 shouldn\'t fail')
-        self.assertTrue(len(r.stdout.split('\n')) == 2)
+        self.assertTrue(len(get_lines(r.stdout)) == 2)
 
         r = rpc('tree', cik, '--level=2')
         self.ok(r, 'tree with --level=2 shouldn\'t fail')
-        self.assertTrue(len(r.stdout.split('\n')) == len(stdports) + 1 + 1)
+        self.assertTrue(len(get_lines(r.stdout)) == len(stdports) + 1 + 1)
 
     def map_test(self):
         '''Map/unmap commands'''
@@ -483,7 +485,6 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         dataports = {}
         resources = [
             Resource(cik, 'dataport', {'format': 'integer', 'name': 'int_port'}),
-            Resource(cik, 'dataport', {'format': 'boolean', 'name': 'boolean_port'}),
             Resource(cik, 'dataport', {'format': 'string', 'name': 'string_port'}),
             Resource(cik, 'dataport', {'format': 'float', 'name': 'float_port'}),
         ]
@@ -608,7 +609,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         self.assertEqual(readscript(childcik3, lua1['name']), lua2['content'], "grandchild updated")
 
 
-    def usage_test(self):
+    def usage_t(self):
         '''Get resource usage'''
         # This test passes inconsistently due to time passing between calls to
         # usage. Mainly all it was testing was date parsing, though.
@@ -682,7 +683,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
             local_time_without_offset = datetime.fromtimestamp(parse_ts(s))
             return str(tz.localize(local_time_without_offset))
 
-        self.ok(r, 'two readings on one timestamp', match='{0},,3,0.3'.format(tolocaltz('2013-07-20 03:00:08')))
+        self.ok(r, 'two readings on one timestamp', match=re.escape('{0},,3,0.3'.format(tolocaltz('2013-07-20 03:00:08'))))
 
         r = rpc('read', '--start=2013-07-20T2:40:07', '--end=2013-07-20T2:40:09', cik, *rids)
         self.l(r.stdout)
@@ -706,6 +707,71 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         r = rpc('read', '--start=2013-07-20T3:00:08', '--end=2013-07-20T3:00:08', '--timeformat=unix', cik)
         self.ok(r, 'read all RIDs', match='timestamp,' + ','.join(['.*' for rid in rids]) + '\n[0-9]+' + ','.join(['.*' for rid in rids]))
 
+    def clone_test(self):
+        '''Clone command'''
+        stdports = self._createDataports()
+        cik = self.client.cik()
+
+        r = rpc('create', cik, '--type=dataport', '--format=string', '--alias=foo')
+        self.ok(r, 'create dataport succeeds')
+        r = rpc('write', cik, 'foo', '--value=testvalue')
+
+        def clone_helper(cik, rid, nohistorical=False):
+            if nohistorical:
+                r = rpc('clone', cik, '--rid=' + rid, '--nohistorical')
+            else:
+                r = rpc('clone', cik, '--rid=' + rid)
+            self.ok(r, 'clone succeeds')
+            if 'cik' in r.stdout:
+                copyrid, copycik = self._ridcik(r.stdout)
+                return copyrid, copycik
+            else:
+                return copyrid
+
+        # clone client
+        copyrid, copycik = clone_helper(self.portalcik, self.client.rid)
+
+        r = rpc('diff', cik, copycik)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff with clone, no differences', match='')
+
+        r = rpc('read', copycik, 'foo', '--format=raw')
+        self.ok(r, 'time series data was copied', match='testvalue')
+
+        copyrid, copycik = clone_helper(self.portalcik, self.client.rid, nohistorical=True)
+
+        r = rpc('diff', cik, copycik)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff with clone, no differences', match='')
+
+        r = rpc('read', copycik, 'foo', '--format=raw')
+        self.ok(r, 'time series data was not copied when --nohistorical was specified', match='')
+
+        # clone dataport
+        '''copyrid = clone_helper(cik, 'foo')
+        self.ok(r, 'copy a dataport')
+
+        r = rpc('diff', cik, copycik)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff after cloning dataport in only one of two clones, notices differences', match='.+')
+
+        copyrid = clone_helper(copycik, 'foo')
+        self.ok(r, 'copy another dataport')
+
+        r = rpc('diff', cik, copycik)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff after cloning dataport in two clones, no differences', match='')'''
+
+        # TODO: test --noaliases
+
     def copy_diff_test(self):
         '''Copy and diff commands'''
         stdports = self._createDataports()
@@ -1051,6 +1117,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         '''Spec command'''
         cik = self.client.cik()
         spec = 'files/spec.yaml'
+        spec_schema = 'files/spec_schema.yaml'
         r = rpc('spec', cik, spec)
         self.notok(r, 'no ids specified')
         r = rpc('spec', cik, spec, '--ids=1,2')
@@ -1061,13 +1128,23 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         self.ok(r, 'create dataports and scripts based on spec', match='')
 
         # test correct number of children of each type
+        r = rpc('drop', cik, '--all-children')
+        self.ok(r, 'drop children from client')
+        # just one id so numbers work out
+        r = rpc('spec', cik, spec, '--ids=1', '--create')
+        self.ok(r, 'create dataports and scripts based on spec')
+
         r = rpc('listing', cik, '--types=dataport,datarule')
         self.ok(r, 'get listing of client')
         listing = json.loads(r.stdout)
         dataports = listing['dataport']
         datarules = listing['datarule']
-        self.assertEqual(len(dataports), 5, 'created correct number of dataports')
-        self.assertEqual(len(datarules), 3, 'created correct number of datarules')
+
+        specobj = yaml.safe_load(open(spec))
+        spec_numdataports = len(specobj['dataports'])
+        self.assertEqual(len(dataports), spec_numdataports, 'created correct number of dataports')
+        spec_numscripts= len(specobj['scripts'])
+        self.assertEqual(len(datarules), spec_numscripts, 'created correct number of scripts')
 
         # test initial value
         r = rpc('read', cik, 'teststring_set1', '--format=raw')
@@ -1099,6 +1176,24 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         r = rpc('read', cik, 'testjson', '--format=raw')
         self.ok(r, 'read testjson initial value', match='{}')
 
+        # test json schema validation
+        r = rpc('drop', cik, '--all-children')
+        self.ok(r, 'drop children from test client')
+        r = rpc('create', cik, '--type=dataport', '--format=string', '--alias=testschema')
+        r = rpc('write', cik, 'testschema', "--value={\"foo\":\"bar\"}")
+        self.ok(r, 'write valid json')
+        r = rpc('read', cik, 'testschema')
+        self.l('stdout: ' + r.stdout)
+        r = rpc('spec', cik, spec_schema)
+        self.ok(r, 'jsonschema is valid', match='')
+
+        r = rpc('write', cik, 'testschema', "--value={\"bar\":1}")
+        self.ok(r, 'write invalid json')
+        r = rpc('read', cik, 'testschema')
+        self.l('stdout: ' + r.stdout)
+        r = rpc('spec', cik, spec_schema)
+        self.ok(r, 'jsonschema is not valid', search='is a required property')
+
         # TODO: test lua script templating
         # TODO: test that correct differences are reported. Change one thing
         # at a time, confirm report is correct, then confirm passing --create
@@ -1125,6 +1220,80 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
 
         os.remove(example_spec)
 
+    def spec_multi_test(self):
+        '''Test spec --portal for updating multiple devices'''
+        # Get example spec
+        r = rpc('spec', '--example')
+        example_spec = 'files/tmp_examplespec.yaml'
+        with open(example_spec, 'w') as f:
+            print(r.stdout)
+            if six.PY3:
+                out = r.stdout
+            else:
+                out = r.stdout.encode('utf8')
+            f.write(out)
+
+
+        cik = self.client.cik()
+
+        # meta fields for test devices
+        metaMyModel = "{\"device\":{\"type\":\"vendor\",\"model\":\"myModel\",\"vendor\":\"myVendor\"}}"
+        metaNotMyModel = "{\"device\":{\"type\":\"vendor\",\"model\":\"NotMyModel\",\"vendor\":\"myVendor\"}}"
+
+
+        # Create two devices of myModel type,
+        myDev1_r = Resource(
+            cik,
+            'client',
+            {"name": "myDev1",
+            "meta":metaMyModel})
+
+        myDev2_r = Resource(
+            cik,
+            'client',
+            {"name": "myDev2",
+            "meta":metaMyModel})
+
+        # Create one device of notMyModel type
+        notMyDev_r = Resource(
+            cik,
+            'client',
+            {"name": "notMyDev",
+            "meta":metaNotMyModel})
+
+        # Create once device without a model type
+        genericDev_r = Resource(
+            cik,
+            'client',
+            {"name": "genericDev"})
+
+
+        # Create devices
+        myDev1 = self._create(myDev1_r)
+        myDev2 = self._create(myDev2_r)
+        notMyDev = self._create(notMyDev_r)
+        genericDev = self._create(genericDev_r)
+
+        # Attempt to apply spec to myModel types
+        r = rpc('spec', cik, example_spec, '--portal', '-f', '--create', '--update-scripts', '--ids=A,B')
+        self.ok(r, 'applying spec to portal')
+
+        # make sure that both myDevs now meet spec
+        r = rpc('spec', myDev1.cik(), example_spec, '--ids=A,B')
+        self.ok(r, "Device 1 didn't match spec", search='')
+
+        r = rpc('spec', myDev2.cik(), example_spec, '--ids=A,B')
+        self.ok(r, "Device 2 didn't match spec", search='')
+
+        # and that both the non-example or the one that didn't have
+        # a type don't meet the spec.
+        r = rpc('spec', notMyDev.cik(), example_spec, '--ids=A,B')
+        self.ok(r, "Device didn't match spec", search='not found')
+
+        r = rpc('spec', genericDev.cik(), example_spec, '--ids=A,B')
+        self.ok(r, "Device didn't match spec", search='not found')
+
+
     def portals_cache_test(self):
         '''Portals clearcache command and option'''
         cik = self.client.cik()
@@ -1260,3 +1429,44 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
                 childcik1,
                 '--share=' + share_code)
         self.ok(r, 'look up RID for share_code', match=dataport_rid1)
+
+
+    def flush_test(self):
+        '''Flush command'''
+        cik = self.client.cik()
+
+        dp1 = self._createMultiple(
+            cik,
+            [Resource(cik, 'dataport',
+                      {'name': 'dp1',
+                       'format': 'string'},
+                      alias='dp1')])
+
+        r = rpc('record', cik, 'dp1', '--value=1391205573,str1', '--value=1391206208,str2', '--value=1391206232,str3')
+        self.ok(r, 'record some values')
+        r = rpc('flush', cik, 'dp1')
+        self.ok(r, 'flush all points')
+        r = rpc('read', cik, 'dp1', '--limit=3', '--sort=asc')
+        self.ok(r, 'points were flushed', match='')
+
+        r = rpc('record', cik, 'dp1', '--value=1391205573,str1', '--value=1391206208,str2', '--value=1391206232,str3')
+        self.ok(r, 'record values')
+        r = rpc('flush', cik, 'dp1', '--start=1391206208')
+        self.ok(r, 'flush with a start time')
+        r = rpc('read', cik, 'dp1', '--limit=3', '--format=raw', '--sort=asc')
+        self.ok(r, 'points were flushed', match='str1\nstr2')
+
+        r = rpc('record', cik, 'dp1', '--value=1391206232,str3')
+        self.ok(r, 'record values')
+        r = rpc('flush', cik, 'dp1', '--end=1391206208')
+        self.ok(r, 'flush with a end time')
+        r = rpc('read', cik, 'dp1', '--limit=3', '--format=raw', '--sort=asc')
+        self.ok(r, 'points were flushed', match='str2\nstr3')
+
+        r = rpc('record', cik, 'dp1', '--value=1391205573,str1')
+        self.ok(r, 'record values')
+        r = rpc('flush', cik, 'dp1', '--start=1391206207', '--end=1391206233')
+        self.ok(r, 'flush with both start and end time')
+        r = rpc('read', cik, 'dp1', '--limit=3', '--format=raw', '--sort=asc')
+        self.ok(r, 'points were flushed', match='str1')
+
